/* 
 *   Creation Date: <2002/06/08 21:01:54 samuel>
 *   Time-stamp: <2002/10/29 12:26:14 samuel>
 *   
 *	<fault.c>
 *	
 *	Linux part
 *   
 *   Copyright (C) 2002 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "archinclude.h"
#include "alloc.h"
#include "kernel_vars.h"
#include "mmu.h"
#include "mmu_contexts.h"
#include "asmfuncs.h"
#include "emu.h"
#include "misc.h"
#include "rvec.h"
#include "performance.h"

/* stolen from include/asm-ppc/pgtable.h */
static inline ulong
update_pte(ulong *p, ulong set)
{
	unsigned long old, tmp;
	
	__asm__ __volatile__("\
1:	lwarx	%0,0,%3\n\
	or	%1,%0,%4\n\
	stwcx.	%1,0,%3\n\
	bne-	1b"
	: "=&r" (old), "=&r" (tmp), "=m" (*p)
	: "r" (p), "r" (set), "m" (*p)
	: "cc" );
	return old;
}

/*
 * Get physical page corresponding to linux virtual address.
 * Invokes linux page fault handler if the page is missing.
 * This function never fails since we know there is a
 * valid mapping for va.
 */

#ifdef NEW_LINUX_MM
 #define PAGE_BITS_WRITE	(_PAGE_ACCESSED | _PAGE_DIRTY | _PAGE_HASHPTE )
 #define PAGE_BITS_READ		(_PAGE_ACCESSED | _PAGE_HASHPTE )
#else
 #define PAGE_BITS_WRITE	(_PAGE_ACCESSED | _PAGE_DIRTY | _PAGE_HWWRITE )
 #define PAGE_BITS_READ		(_PAGE_ACCESSED )
#endif

ulong 
get_phys_page( ulong va, int request_rw )
{
	ulong	val, uptr, *ptr;
	ulong	flags;
	struct	mm_struct *mm;
	struct	vm_area_struct *vma;
	flags = request_rw ? 
		(_PAGE_USER | _PAGE_RW | _PAGE_PRESENT)
		: (_PAGE_USER | _PAGE_PRESENT );
	
#ifdef LINUX_24
	uptr = ((ulong*)current->thread.pgdir)[va>>22];	/* top 10 bits */
#else
	uptr = current->tss.pg_tables[va>>22];	/* top 10 bits */
#endif
	ptr = (ulong*)( uptr & 0xfffff000 );
	if( !ptr )
		goto no_page;
	
	ptr = ptr + ((va>>12) & 0x3ff);	      	/* next 10 bits */
	val = *ptr;
	if( flags & ~val )			/* check access priv */
		goto no_page;

	// we set _PAGE_HASHPTE in order to ensure that we get to
	// know about TLB invalidations for this page via our
	// flush_hash_page hook.
	update_pte(ptr, (request_rw? PAGE_BITS_WRITE : PAGE_BITS_READ ));
	// printk("pte at %p was %lx now %lx\n", ptr, val, *ptr);

	return val & ~0xfff;

no_page:
	BUMP( page_missing );

	// no mac page found...
	mm = CURRENT_MM;
	MMAP_SEM_DOWN( &mm->mmap_sem );

	if( !(vma=find_vma( mm,va )) || vma->vm_start > va )
		goto bad_area;
	if( !(vma->vm_flags & (request_rw ? VM_WRITE : (VM_READ | VM_EXEC))) )
		goto bad_area;

#ifdef LINUX_24
	mc_handle_mm_fault( current->mm, vma, va, request_rw );
#else
	mc_handle_mm_fault( current, vma, va, request_rw );
#endif
	MMAP_SEM_UP( &mm->mmap_sem );
	return get_phys_page(va, request_rw);

bad_area:
	MMAP_SEM_UP( &mm->mmap_sem );
	printk("get_phys_page: BAD AREA, lvptr = %08lX\n",va );
	force_sig(SIGSEGV, current);
	return 0;
}


/************************************************************************/
/*	Debugger functions						*/
/************************************************************************/

int 
dbg_get_linux_page( ulong va, linux_page_t *r )
{
	ulong	val, uptr, *ptr;
	
	if( verify_area(VERIFY_WRITE, r, sizeof(*r)) )
		return -EFAULT;

#ifdef LINUX_24
	uptr = ((ulong*)current->thread.pgdir)[va>>22];	/* top 10 bits */
#else
	uptr = current->tss.pg_tables[va>>22];	/* top 10 bits */
#endif
	ptr = (ulong*)( uptr & 0xfffff000 );
	if( !ptr )
		return 1;
	val = ptr[ (va>>12)&0x3ff ];		/* next 10 bits */

	r->phys = val & ~0xfff;
	r->mflags = 
		  DBG_TRANSL_PAGE_FLAG( val, _PAGE_PRESENT )
		| DBG_TRANSL_PAGE_FLAG( val, _PAGE_USER )
		| DBG_TRANSL_PAGE_FLAG(	val, _PAGE_GUARDED )
		| DBG_TRANSL_PAGE_FLAG( val, _PAGE_COHERENT )
		| DBG_TRANSL_PAGE_FLAG( val, _PAGE_NO_CACHE )
		| DBG_TRANSL_PAGE_FLAG( val, _PAGE_WRITETHRU )
		| DBG_TRANSL_PAGE_FLAG( val, _PAGE_DIRTY )
		| DBG_TRANSL_PAGE_FLAG( val, _PAGE_ACCESSED )
		| DBG_TRANSL_PAGE_FLAG( val, _PAGE_RW )
		| DBG_TRANSL_PAGE_FLAG_N( val, _PAGE_HASHPTE )
		| DBG_TRANSL_PAGE_FLAG_N( val, _PAGE_EXEC );
	return 0;
}
